/* @(#)isoinfo.c	1.108 16/08/10 joerg */
#include <schily/mconfig.h>
/*
 * File isodump.c - dump iso9660 directory information.
 *
 *
 * Written by Eric Youngdale (1993).
 *
 * Copyright 1993 Yggdrasil Computing, Incorporated
 * Copyright (c) 1999-2016 J. Schilling
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; see the file COPYING.  If not, write to the Free Software
 * Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/*
 * Simple program to dump contents of iso9660 image in more usable format.
 *
 * Usage:
 * To list contents of image (with or without RR):
 *	isoinfo -l [-R] -i imagefile
 * To extract file from image:
 *	isoinfo -i imagefile -x xtractfile > outfile
 * To generate a "find" like list of files:
 *	isoinfo -f -i imagefile
 */

#include <schily/stdlib.h>
#include <schily/unistd.h>
#include <schily/string.h>

#include <schily/stdio.h>
#include <schily/utypes.h>
#include <schily/standard.h>
#include <schily/signal.h>
#include <schily/stat.h>
#include <schily/device.h>
#include <schily/time.h>
#include <schily/fcntl.h>
#include <schily/intcvt.h>
#include <schily/nlsdefs.h>
#include <schily/getargs.h>
#include <schily/nlsdefs.h>
#include <schily/ctype.h>
#include <schily/errno.h>
#include <schily/schily.h>

#include "iso9660.h"
#include "rock.h"
#include "scsi.h"
#include "cdrdeflt/cdrdeflt.h"

#define	VERSION	"3.02a09"

#include <schily/siconv.h>
#include <schily/io.h>				/* for setmode() prototype */

/*
 * Make sure we have a definition for this.  If not, take a very conservative
 * guess.
 * POSIX requires the max pathname component lenght to be defined in limits.h
 * If variable, it may be undefined. If undefined, there should be
 * a definition for _POSIX_NAME_MAX in limits.h or in unistd.h
 * As _POSIX_NAME_MAX is defined to 14, we cannot use it.
 * XXX Eric's wrong comment:
 * XXX From what I can tell SunOS is the only one with this trouble.
 */
#include <schily/limits.h>

#ifndef	NAME_MAX
#ifdef	FILENAME_MAX
#define	NAME_MAX	FILENAME_MAX
#else
#define	NAME_MAX	256
#endif
#endif

#ifndef	PATH_MAX
#ifdef	FILENAME_MAX
#define	PATH_MAX	FILENAME_MAX
#else
#define	PATH_MAX	1024
#endif
#endif

/*
 * XXX JS: Some structures have odd lengths!
 * Some compilers (e.g. on Sun3/mc68020) padd the structures to even length.
 * For this reason, we cannot use sizeof (struct iso_path_table) or
 * sizeof (struct iso_directory_record) to compute on disk sizes.
 * Instead, we use offsetof(..., name) and add the name size.
 * See iso9660.h
 */
#ifndef	offsetof
#define	offsetof(TYPE, MEMBER)	((size_t) &((TYPE *)0)->MEMBER)
#endif

/*
 * Note: always use these macros to avoid problems.
 *
 * ISO_ROUND_UP(X)	may cause an integer overflow and thus give
 *			incorrect results. So avoid it if possible.
 *
 * ISO_BLOCKS(X)	is overflow safe. Prefer this when ever it is possible.
 */
#define	SECTOR_SIZE	(2048)
#define	ISO_ROUND_UP(X)	(((X) + (SECTOR_SIZE - 1)) & ~(SECTOR_SIZE - 1))
#define	ISO_BLOCKS(X)	(((X) / SECTOR_SIZE) + (((X)%SECTOR_SIZE)?1:0))

#define	GET_UBYTE(a)	a_to_u_byte(a)

struct iso_extract_handle {
	struct scsi_dev *dev;
	struct iso_directory_record *idr;
};

typedef int(*iso_listing_callback)(void*,const char*,int is_dir,int sz,
		struct iso_extract_handle*);

enum ScanRetCode {
	ScanRetCode_Ok = 100,
	ScanRetCode_Disabled,
	ScanRetCode_BadInput
};

EXPORT BOOL	ignerr = FALSE;
LOCAL int	use_rock = 0;
LOCAL int	use_joliet = 0;
LOCAL int	do_f = 0;
LOCAL int	do_sectors = 0;
LOCAL BOOL	debug = FALSE;
LOCAL char	*xtract = 0;
LOCAL BOOL	Xtract = FALSE;
LOCAL char	er_id[256];
LOCAL int	su_version = 0;
LOCAL int	rr_version = 0;
LOCAL int	aa_version = 0;
LOCAL int	cl_extent = 0;
LOCAL int	ucs_level = 0;
LOCAL BOOL	iso9660_inodes = FALSE;
LOCAL uid_t	myuid;

LOCAL struct stat	fstat_buf;
LOCAL int		found_rr;
LOCAL char		name_buf[256*3];
LOCAL char		xname[8192];
LOCAL unsigned char	date_buf[9];	/* iso_directory_record.date[7] */
/*
 * Use sector_offset != 0 (-N #) if we have an image file
 * of a single session and we need to list the directory contents.
 * This is the session block (sector) number of the start
 * of the session when it would be on disk.
 */
LOCAL unsigned int	sector_offset = 0;

LOCAL unsigned char	buffer[2048];

LOCAL siconvt_t	*unls;

#define	PAGE sizeof (buffer)

#define	ISODCL(from, to) (to - from + 1)

struct todo
{
	struct todo	*next;
	struct todo	*prev;
	char		*name;
	int		extent;
	int		length;
};

LOCAL struct todo	*todo_idr = NULL;
LOCAL struct todo	**todo_pp = &todo_idr;


LOCAL	int	isonum_731	__PR((char * p));
LOCAL	int	isonum_733	__PR((unsigned char * p));
LOCAL	int	parse_rr	__PR((struct scsi_dev *dev, unsigned char * pnt,
					int len, int cont_flag));
LOCAL	void	find_rr		__PR((struct iso_directory_record * idr,
					Uchar **pntp, int *lenp));
LOCAL	int	dump_rr		__PR((struct scsi_dev *dev,
					struct iso_directory_record * idr));
LOCAL	BOOL	dump_stat	__PR((char *rootname,
					struct iso_directory_record * idr,
					char *fname,
					int extent));
LOCAL	void	extract		__PR((struct scsi_dev *dev,
					char *rootname, struct iso_directory_record * idr,
					char *fname));
LOCAL	void	extract_file	__PR((struct scsi_dev *dev, int f,
					struct iso_directory_record * idr,
					char *fname));
EXPORT void extract_file_from_iso __PR((struct iso_extract_handle *context,
					char *out));
LOCAL	void	parse_cl_dir	__PR((struct scsi_dev *dev,
					struct iso_directory_record *idr, int extent));
LOCAL	BOOL	parse_de	__PR((struct iso_directory_record *idr));
LOCAL	int parse_dir __PR((struct scsi_dev *dev, struct todo *dp,
					char * rootname, int extent, int len,
					iso_listing_callback callback, void* user_data));
LOCAL	void	list_locales	__PR((void));
LOCAL	int	time_cvt	__PR((unsigned char *dp, int len));
LOCAL	time_t	iso9660_time	__PR((unsigned char *date, int *hsecp,
					BOOL longfmt));
LOCAL	int run_scanning __PR((struct scsi_dev *dev,
					iso_listing_callback callback, void *user_data));
LOCAL char* default_charset __PR((void));

LOCAL int
isonum_731(p)
	char	*p;
{
	return ((p[0] & 0xff)
		| ((p[1] & 0xff) << 8)
		| ((p[2] & 0xff) << 16)
		| ((p[3] & 0xff) << 24));
}

LOCAL int
isonum_733(p)
	unsigned char	*p;
{
	return (isonum_731((char *)p));
}

#ifdef	USE_ICONV
#ifdef	HAVE_ICONV_CONST
#define	__IC_CONST	const
#else
#define	__IC_CONST
#endif
#endif

LOCAL int
parse_rr(dev, pnt, len, cont_flag)
	struct scsi_dev *dev;
	unsigned char	*pnt;
	int		len;
	int		cont_flag;
{
	int slen;
	int ncount;
	int cont_extent, cont_offset, cont_size;
	int flag2;
	unsigned char *pnts;
	char symlinkname[1024];
	int goof = 0;

	symlinkname[0] = 0;

	cl_extent = cont_extent = cont_offset = cont_size = 0;

	ncount = 0;
	flag2 = 0;
	while (len >= 4) {
		if (pnt[3] != 1 && pnt[3] != 2) {
			printf(_("**BAD RRVERSION (%d) in '%2.2s' field %2.2X %2.2X.\n"), pnt[3], pnt, pnt[0], pnt[1]);
			return (0);		/* JS ??? Is this right ??? */
		}
		if (pnt[2] < 4) {
			printf(_("**BAD RRLEN (%d) in '%2.2s' field %2.2X %2.2X.\n"), pnt[2], pnt, pnt[0], pnt[1]);
			return (0);		/* JS ??? Is this right ??? */
		}

		ncount++;
		if (strncmp((char *)pnt, "PX", 2) == 0) flag2 |= RR_FLAG_PX;	/* POSIX attributes */
		if (strncmp((char *)pnt, "PN", 2) == 0) flag2 |= RR_FLAG_PN;	/* POSIX device number */
		if (strncmp((char *)pnt, "SL", 2) == 0) flag2 |= RR_FLAG_SL;	/* Symlink */
		if (strncmp((char *)pnt, "NM", 2) == 0) flag2 |= RR_FLAG_NM;	/* Alternate Name */
		if (strncmp((char *)pnt, "CL", 2) == 0) flag2 |= RR_FLAG_CL;	/* Child link */
		if (strncmp((char *)pnt, "PL", 2) == 0) flag2 |= RR_FLAG_PL;	/* Parent link */
		if (strncmp((char *)pnt, "RE", 2) == 0) flag2 |= RR_FLAG_RE;	/* Relocated Direcotry */
		if (strncmp((char *)pnt, "TF", 2) == 0) {
			BOOL	longfmt;
			int	size = 7;
			int	hsec;
			unsigned char *p = &pnt[5];

			flag2 |= RR_FLAG_TF;					/* Time stamp */
			longfmt = (pnt[4] & 0x80) != 0;
			if (longfmt)
				size = 17;
			if (pnt[4] & 0x01) {
				p += size;
			}
			if (pnt[4] & 0x02) {
				fstat_buf.st_mtime = iso9660_time(p,
							&hsec, longfmt);
				hsec *= 10000000;
				stat_set_mnsecs(&fstat_buf, hsec);
				p += size;
			}
			if (pnt[4] & 0x04) {
				fstat_buf.st_atime = iso9660_time(p,
							&hsec, longfmt);
				hsec *= 10000000;
				stat_set_ansecs(&fstat_buf, hsec);
				p += size;
			}
			if (pnt[4] & 0x08) {
				fstat_buf.st_ctime = iso9660_time(p,
							&hsec, longfmt);
				hsec *= 10000000;
				stat_set_cnsecs(&fstat_buf, hsec);
				p += size;
			}
		}
		if (strncmp((char *)pnt, "SF", 2) == 0) flag2 |= RR_FLAG_SF;	/* Sparse File */

		if (strncmp((char *)pnt, "SP", 2) == 0) {
			flag2 |= RR_FLAG_SP;					/* SUSP record */
			su_version = pnt[3] & 0xff;
		}
		if (strncmp((char *)pnt, "AA", 2) == 0) {			/* Neither SUSP nor RR */
			flag2 |= RR_FLAG_AA;					/* Apple Signature record */
			aa_version = pnt[3] & 0xff;
		}
		if (strncmp((char *)pnt, "ER", 2) == 0) {
			flag2 |= RR_FLAG_ER;					/* ER record */
			rr_version = pnt[7] & 0xff;				/* Ext Version */
			strlcpy(er_id, (char *)&pnt[8], (pnt[4] & 0xFF) + 1);
		}

		if (strncmp((char *)pnt, "PX", 2) == 0) {			/* POSIX attributes */
			fstat_buf.st_mode = isonum_733(pnt+4);
			fstat_buf.st_nlink = isonum_733(pnt+12);
			fstat_buf.st_uid = isonum_733(pnt+20);
			fstat_buf.st_gid = isonum_733(pnt+28);
			if ((pnt[2] & 0xFF) >= 44)				/* Check for RR 1.12 */
				fstat_buf.st_ino = (UInt32_t)isonum_733(pnt+36);
		}

		if (strncmp((char *)pnt, "PN", 2) == 0) {			/* POSIX device number */
			dev_t	devmajor = isonum_733(pnt+4);
			dev_t	devminor = isonum_733(pnt+12);

			fstat_buf.st_rdev = makedev(devmajor, devminor);
		}

		if (strncmp((char *)pnt, "NM", 2) == 0) {		/* Alternate Name */
			int	l = strlen(name_buf);

			if (!found_rr)
				l = 0;
			strncpy(&name_buf[l], (char *)(pnt+5), pnt[2] - 5);
			name_buf[l + pnt[2] - 5] = 0;
			found_rr = 1;
		}

		if (strncmp((char *)pnt, "CE", 2) == 0) {		/* Continuation Area */
			cont_extent = isonum_733(pnt+4);
			cont_offset = isonum_733(pnt+12);
			cont_size = isonum_733(pnt+20);
		}

		if (strncmp((char *)pnt, "ST", 2) == 0) {		/* Terminate SUSP */
			break;
		}

		if (strncmp((char *)pnt, "CL", 2) == 0) {
			cl_extent = isonum_733(pnt+4);			/* Child link location */
		}

		if (strncmp((char *)pnt, "SL", 2) == 0) {		/* Symlink */
			pnts = pnt+5;
			slen = pnt[2] - 5;
			while (slen >= 1) {
				switch (pnts[0] & 0xfe) {
				case 0:
					strncat(symlinkname, (char *)(pnts+2), pnts[1]);
					symlinkname[pnts[1]] = 0;
					break;
				case 2:
					strcat(symlinkname, ".");
					break;
				case 4:
					strcat(symlinkname, "..");
					break;
				case 8:
					symlinkname[0] = '\0';
					break;
				case 16:
					strcat(symlinkname, "/mnt");
					printf(_("Warning - mount point requested"));
					break;
				case 32:
					strcat(symlinkname, "kafka");
					printf(_("Warning - host_name requested"));
					break;
				default:
					printf(_("Reserved bit setting in symlink"));
					goof++;
					break;
				}
				if ((pnts[0] & 0xfe) && pnts[1] != 0) {
					printf(_("Incorrect length in symlink component"));
				}
				if (pnts[0] == 8)
					xname[0] = '\0';
				if (xname[0] == 0)
					strcpy(xname, "-> ");
				strcat(xname, symlinkname);
				symlinkname[0] = 0;
				if ((pnts[0] & 1) == 0)
					strcat(xname, "/");

				slen -= (pnts[1] + 2);
				pnts += (pnts[1] + 2);
			}
			symlinkname[0] = 0;
		}

		len -= pnt[2];
		pnt += pnt[2];
	}
	if (cont_extent) {
		unsigned char	sector[2048];

#ifdef	USE_SCG
		readsecs(dev, cont_extent - sector_offset, sector, ISO_BLOCKS(sizeof (sector)));
#else
		lseek(fileno(infile), ((off_t)(cont_extent - sector_offset)) << 11, SEEK_SET);
		read(fileno(infile), sector, sizeof (sector));
#endif
		flag2 |= parse_rr(dev, &sector[cont_offset], cont_size, 1);
	}
	/*
	 * for symbolic links, strip out the last '/'
	 */
	if (xname[0] != 0 && xname[strlen(xname)-1] == '/') {
		if (strlen(xname) > 4)
			xname[strlen(xname)-1] = '\0';
	}
	return (flag2);
}

LOCAL void
find_rr(idr, pntp, lenp)
	struct iso_directory_record *idr;
	Uchar	**pntp;
	int	*lenp;
{
	struct iso_xa_dir_record *xadp;
	int len;
	unsigned char * pnt;

	len = idr->length[0] & 0xff;
	len -= offsetof(struct iso_directory_record, name[0]);
	len -= idr->name_len[0];

	pnt = (unsigned char *) idr;
	pnt += offsetof(struct iso_directory_record, name[0]);
	pnt += idr->name_len[0];
	if ((idr->name_len[0] & 1) == 0) {
		pnt++;
		len--;
	}
	if (len >= 14) {
		xadp = (struct iso_xa_dir_record *)pnt;

		if (xadp->signature[0] == 'X' && xadp->signature[1] == 'A' &&
		    xadp->reserved[0] == '\0') {
			len -= 14;
			pnt += 14;
		}
	}
	*pntp = pnt;
	*lenp = len;
}

LOCAL int
dump_rr(dev, idr)
	struct scsi_dev *dev;
	struct iso_directory_record *idr;
{
	int len;
	unsigned char * pnt;

	find_rr(idr, &pnt, &len);
	return (parse_rr(dev, pnt, len, 0));
}

LOCAL char		*months[12] = {"Jan", "Feb", "Mar", "Apr",
				"May", "Jun", "Jul",
				"Aug", "Sep", "Oct", "Nov", "Dec"};

/*
 * Return TRUE if this file was "selected" either based on find rules
 * or for all files if there was no find rule.
 */
LOCAL BOOL
dump_stat(rootname, idr, fname, extent)
	char	*rootname;
	struct iso_directory_record *idr;
	char	*fname;
	int	extent;
{
	int	i;
	int	off = 0;
	char	outline[100];

	memset(outline, ' ', sizeof (outline));

	if (fstat_buf.st_ino != 0)
		off = js_sprintf(outline, "%10llu ", (ULlong)fstat_buf.st_ino);
	if (off < 0)
		off = 0;

	if (S_ISREG(fstat_buf.st_mode))
		outline[off] = '-';
	else if (S_ISDIR(fstat_buf.st_mode))
		outline[off] = 'd';
	else if (S_ISLNK(fstat_buf.st_mode))
		outline[off] = 'l';
	else if (S_ISCHR(fstat_buf.st_mode))
		outline[off] = 'c';
	else if (S_ISBLK(fstat_buf.st_mode))
		outline[off] = 'b';
	else if (S_ISFIFO(fstat_buf.st_mode))
		outline[off] = 'f';
	else if (S_ISSOCK(fstat_buf.st_mode))
		outline[off] = 's';
	else
		outline[off] = '?';

	memset(outline+off+1, '-', 9);
	if (fstat_buf.st_mode & S_IRUSR)
		outline[off+1] = 'r';
	if (fstat_buf.st_mode & S_IWUSR)
		outline[off+2] = 'w';
	if (fstat_buf.st_mode & S_IXUSR)
		outline[off+3] = 'x';

	if (fstat_buf.st_mode & S_IRGRP)
		outline[off+4] = 'r';
	if (fstat_buf.st_mode & S_IWGRP)
		outline[off+5] = 'w';
	if (fstat_buf.st_mode & S_IXGRP)
		outline[off+6] = 'x';

	if (fstat_buf.st_mode & S_IROTH)
		outline[off+7] = 'r';
	if (fstat_buf.st_mode & S_IWOTH)
		outline[off+8] = 'w';
	if (fstat_buf.st_mode & S_IXOTH)
		outline[off+9] = 'x';

	off += 10;
	off += js_sprintf(outline+off, " %3ld", (long)fstat_buf.st_nlink);
	off += js_sprintf(outline+off, " %4ld", (unsigned long)fstat_buf.st_uid);
	off += js_sprintf(outline+off, " %4ld", (unsigned long)fstat_buf.st_gid);

	if (do_sectors == 0) {
		off += js_sprintf(outline+off, " %10lld", (Llong)fstat_buf.st_size);
	} else {
		off += js_sprintf(outline+off, " %10lld", (Llong)((fstat_buf.st_size+PAGE-1)/PAGE));
	}

	if (date_buf[1] >= 1 && date_buf[1] <= 12) {
		memcpy(outline+off+1, months[date_buf[1]-1], 3);
		off += 4;
	} else {
		/*
		 * Some broken ISO-9660 formatters write illegal month numbers.
		 */
		off += 1 + js_sprintf(outline+off+1, "???");
	}

	off += js_sprintf(outline+off, " %2d", date_buf[2]);
	off += js_sprintf(outline+off, " %4d", date_buf[0]+1900);

	off += js_sprintf(outline+off, " [%7d", extent);	/* XXX up to 20 GB */
	off += js_sprintf(outline+off, " %02X]", idr->flags[0] & 0xFF);

	for (i = 0; i < off; i++) {
		if (outline[i] == 0) outline[i] = ' ';
	}
	outline[off] = 0;
	printf("%s %s %s\n", outline, name_buf, xname);
	return (TRUE);
}

LOCAL void
extract(dev, rootname, idr, fname)
	struct scsi_dev *dev;
	char				*rootname;
	struct iso_directory_record	*idr;
	char				*fname;
{
	int		f;
	struct timespec	times[2];
static	BOOL		isfirst = TRUE;

	if ((idr->flags[0] & 2) != 0 &&
	    (idr->name_len[0] == 1 &&
	    (idr->name[0] == 0 || idr->name[0] == 1))) {
		/*
		 * Catch all "." and ".." entries here.
		 */
		if (idr->name[0] == 1)	/* Skip "/.." */
			return;
		if (rootname[0] == '/' && rootname[1] == '\0')
			fname = "/.";
	}
	if (*fname == '/')
		fname++;
	makedirs(fname, S_IRUSR|S_IWUSR|S_IXUSR|
			S_IRGRP|S_IWGRP|S_IXGRP|
			S_IROTH|S_IWOTH|S_IXOTH, TRUE);

	switch (fstat_buf.st_mode & S_IFMT) {

#ifndef	HAVE_MKNOD
	err:
#endif
	default:
			errmsgno(EX_BAD,
				"Unsupported file type %0lo for '%s'.\n",
				(unsigned long)fstat_buf.st_mode & S_IFMT,
				fname);
			return;


	case S_IFDIR:
#ifdef	__MINGW32__
#define	mkdir(n, m)	mkdir(n)
#endif
			if (mkdir(fname, fstat_buf.st_mode) < 0 &&
			    geterrno() != EEXIST)
				errmsg("Cannot make directory '%s'.\n", fname);
			goto setmode;

#ifdef	S_IFBLK
	case S_IFBLK:
#ifdef	HAVE_MKNOD
			if (mknod(fname, fstat_buf.st_mode, fstat_buf.st_rdev) < 0)
				errmsg("Cannot make block device '%s'.\n", fname);
			goto setmode;
#else
			goto err;
#endif
#endif
#ifdef	S_IFCHR
	case S_IFCHR:
#ifdef	HAVE_MKNOD
			if (mknod(fname, fstat_buf.st_mode, fstat_buf.st_rdev) < 0)
				errmsg("Cannot make character device '%s'.\n", fname);
			goto setmode;
#else
			goto err;
#endif
#endif
#ifdef	S_IFIFO
	case S_IFIFO:
#ifdef	HAVE_MKFIFO
			if (mkfifo(fname, fstat_buf.st_mode) < 0)
				errmsg("Cannot make fifo '%s'.\n", fname);
			goto setmode;
#else
			goto err;
#endif
#endif
#ifdef	S_IFSOCK
	case S_IFSOCK:
#ifdef	HAVE_MKNOD
			if (mknod(fname, fstat_buf.st_mode, 0) < 0)
				errmsg("Cannot make socket '%s'.\n", fname);
			goto setmode;
#else
			goto err;
#endif
#endif
#ifdef	S_IFLNK
	case S_IFLNK:
			if (symlink(&xname[3], fname) < 0)
				errmsg("Cannot make symlink '%s'.\n", fname);
			goto setmode;
#endif

	case S_IFREG:	break;
	}

	makedirs(fname, S_IRUSR|S_IWUSR|S_IXUSR|
			S_IRGRP|S_IWGRP|S_IXGRP|
			S_IROTH|S_IWOTH|S_IXOTH, TRUE);
	if (isfirst)
		f = open(fname, O_WRONLY|O_CREAT|O_TRUNC, 0600);
	else
		f = open(fname, O_WRONLY|O_CREAT, 0600);
	if (f < 0) {
		errmsg("Cannot create '%s'.\n", fname);
		return;
	}
	lseek(f, 0, SEEK_END);
	extract_file(dev, f, idr, fname);
	if ((idr->flags[0] & ISO_MULTIEXTENT) == 0) {
#ifdef	HAVE_FCHOWN
		fchown(f, fstat_buf.st_uid, fstat_buf.st_gid);
#else
		fchownat(AT_FDCWD, fname, fstat_buf.st_uid, fstat_buf.st_gid, 0);
#endif
#ifdef	HAVE_FCHMOD
		fchmod(f, fstat_buf.st_mode);
#else
		fchmodat(AT_FDCWD, fname, fstat_buf.st_mode, 0);
#endif
		times[0].tv_sec = fstat_buf.st_atime;
		times[0].tv_nsec = stat_ansecs(&fstat_buf);
		times[1].tv_sec = fstat_buf.st_mtime;
		times[1].tv_nsec = stat_mnsecs(&fstat_buf);
#if defined(HAVE_FUTIMESAT) || defined(HAVE_FUTIMES) || defined(HAVE_FUTIMENS)
		futimens(f, times);
#else
		utimensat(AT_FDCWD, fname, times, 0);
#endif
		isfirst = TRUE;		/* Next call is for a new file */
	} else {
		isfirst = FALSE;	/* Next call is a continuation */
	}
	close(f);
	return;
setmode:
	fchownat(AT_FDCWD, fname, fstat_buf.st_uid, fstat_buf.st_gid, AT_SYMLINK_NOFOLLOW);
	fchmodat(AT_FDCWD, fname, fstat_buf.st_mode, AT_SYMLINK_NOFOLLOW);
	times[0].tv_sec = fstat_buf.st_atime;
	times[0].tv_nsec = stat_ansecs(&fstat_buf);
	times[1].tv_sec = fstat_buf.st_mtime;
	times[1].tv_nsec = stat_mnsecs(&fstat_buf);
	utimensat(AT_FDCWD, fname, times, AT_SYMLINK_NOFOLLOW);
}

LOCAL void
extract_file(dev, f, idr, fname)
	struct scsi_dev *dev;
	int				f;
	struct iso_directory_record	*idr;
	char				*fname;
{
	int		extent, len, tlen;
	unsigned char	buff[20480];

	if (f == STDOUT_FILENO)
		setmode(fileno(stdout), O_BINARY);

	extent = isonum_733((unsigned char *)idr->extent);
	len = isonum_733((unsigned char *)idr->size);

	while (len > 0) {
		tlen = (len > sizeof (buff) ? sizeof (buff) : len);
#ifdef	USE_SCG
		readsecs(dev, extent - sector_offset, buff, ISO_BLOCKS(tlen));
#else
		lseek(fileno(infile), ((off_t)(extent - sector_offset)) << 11, SEEK_SET);
		read(fileno(infile), buff, tlen);
#endif
		len -= tlen;
		extent += ISO_BLOCKS(tlen);
		if (write(f, buff, tlen) != tlen)
			errmsg("Write error on '%s'.\n", fname);
	}
}


LOCAL void
parse_cl_dir(dev, idr, extent)
	struct scsi_dev *dev;
	struct iso_directory_record	*idr;
	int				extent;
{
	char				cl_name_buf[256*3];

	strlcpy(cl_name_buf, name_buf, sizeof (cl_name_buf));
#ifdef	USE_SCG
	readsecs(dev, extent - sector_offset, idr, 1);
#else
	lseek(fileno(infile), ((off_t)(extent - sector_offset)) << 11, SEEK_SET);
	read(fileno(infile), idr, 2048);
#endif

	if (parse_de(idr) && use_rock)
		dump_rr(dev, idr);
	strlcpy(name_buf, cl_name_buf, sizeof (name_buf));
}

LOCAL BOOL
parse_de(idr)
	struct iso_directory_record	*idr;
{
	unsigned char	uc;

	if (idr->length[0] == 0)
		return (FALSE);
	memset(&fstat_buf, 0, sizeof (fstat_buf));
	found_rr = 0;
	name_buf[0] = xname[0] = 0;
	fstat_buf.st_size = (off_t)(unsigned)isonum_733((unsigned char *)idr->size);
	if (idr->flags[0] & 2)
		fstat_buf.st_mode |= S_IFDIR;
	else
		fstat_buf.st_mode |= S_IFREG;
	if (idr->name_len[0] == 1 && idr->name[0] == 0)
		strcpy(name_buf, ".");
	else if (idr->name_len[0] == 1 && idr->name[0] == 1)
		strcpy(name_buf, "..");
	else {
		switch (ucs_level) {
		case 3:
		case 2:
		case 1:
			/*
			 * Unicode name.  Convert as best we can.
			 */
			{
			int	j;
				name_buf[0] = '\0';
#ifdef	USE_ICONV
			if (use_iconv(unls)) {
				int	u;
				char	*to = name_buf;

				for (j = 0, u = 0; j < (int)idr->name_len[0] / 2; j++) {
					char	*ibuf = (char *)&idr->name[j*2];
					size_t	isize = 2;		/* UCS-2 character size */
					size_t	osize = 4;

					if (iconv(unls->sic_uni2cd, (__IC_CONST char **)&ibuf, &isize,
							(char **)&to, &osize) == -1) {
						int	err = geterrno();

						if ((err == EINVAL || err == EILSEQ) &&
						    osize == 4) {
							*to = '_';
							u += 1;
							to++;
						}
					} else {
						u += 4 - osize;
						to = &name_buf[u];
					}
				}
				j = u;
			} else
#endif
			for (j = 0; j < (int)idr->name_len[0] / 2; j++) {
				UInt16_t	unichar;

				unichar = (idr->name[j*2] & 0xFF) * 256 +
					    (idr->name[j*2+1] & 0xFF);

				/*
				 * Get the backconverted char
				 */
				if (unls)
					uc = sic_uni2c(unls, unichar);
				else
					uc = unichar > 255 ? '_' : unichar;

				name_buf[j] = uc ? uc : '_';
			}
			name_buf[j] = '\0';
			}
			break;
		case 0:
			/*
			 * Normal non-Unicode name.
			 */
			strncpy(name_buf, idr->name, idr->name_len[0]);
			name_buf[idr->name_len[0]] = 0;
			break;
		default:
			/*
			 * Don't know how to do these yet.  Maybe they are the same
			 * as one of the above.
			 */
			exit(1);
		}
	}
	memcpy(date_buf, idr->date, sizeof (idr->date));
	/*
	 * Always first set up time stamps and file modes from
	 * ISO-9660. This is used as a fallback in case that
	 * there is no related Rock Ridge based data.
	 */
	fstat_buf.st_atime =
	fstat_buf.st_mtime =
	fstat_buf.st_ctime = iso9660_time(date_buf, NULL, FALSE);
	fstat_buf.st_mode |= S_IRUSR|S_IXUSR |
		    S_IRGRP|S_IXGRP |
		    S_IROTH|S_IXOTH;
	fstat_buf.st_nlink = 1;
	fstat_buf.st_ino = 0;
	fstat_buf.st_uid = 0;
	fstat_buf.st_gid = 0;
	if (iso9660_inodes) {
		fstat_buf.st_ino = (unsigned long)
		    isonum_733((unsigned char *)idr->extent);
	}
	return (TRUE);
}

LOCAL int
parse_dir(dev, dp, rootname, extent, len, callback, user_data)
	struct scsi_dev *dev;
	struct todo	*dp;
	char	*rootname;
	int	extent;			/* Directory extent */
	int	len;			/* Directory size   */
	iso_listing_callback callback;
	void *user_data;
{
	struct todo 	*td;
	int		i;
	struct iso_directory_record * idr;
	struct iso_extract_handle extr_handle;
	unsigned char	cl_buffer[2048];
	unsigned char	flags = 0;
	Llong		size = 0;
	int		sextent = 0;
	int	rlen;
	int	blen;
	int	rr_flags = 0;
	int stop_now = 0;
static	char	*n = 0;
static	int	nlen = 0;

	rlen = strlen(rootname);

	while (len > 0) {
#ifdef	USE_SCG
		readsecs(dev, extent - sector_offset, buffer, ISO_BLOCKS(sizeof (buffer)));
#else
		lseek(fileno(infile), ((off_t)(extent - sector_offset)) << 11, SEEK_SET);
		read(fileno(infile), buffer, sizeof (buffer));
#endif
		len -= sizeof (buffer);
		extent++;
		i = 0;
		while (1 == 1) {
			idr = (struct iso_directory_record *) &buffer[i];
			if (idr->length[0] == 0)
				break;
			parse_de(idr);
			if (use_rock) {
				rr_flags = dump_rr(dev, idr);

				if (rr_flags & RR_FLAG_CL) {
					/*
					 * Need to reparse the child link
					 * but note that we parse "CL/."
					 * so we get no usable file name.
					 */
					idr = (struct iso_directory_record *) cl_buffer;
					parse_cl_dir(dev, idr, cl_extent);
				} else if (rr_flags & RR_FLAG_RE)
					goto cont;	/* skip rr_moved */
			}
			if (Xtract &&
			    (idr->flags[0] & 2) != 0 &&
			    idr->name_len[0] == 1 &&
			    idr->name[0] == 0) {
				/*
				 * The '.' entry.
				 */
			}
			blen = strlen(name_buf);

			blen = rlen + blen + 1;
			if (nlen < blen) {
				n = ___realloc(n, blen, _("find_stat name"));
				nlen = blen;
			}
			strcatl(n, rootname, name_buf, (char *)0);
			if (name_buf[0] == '.' && name_buf[1] == '\0')
				n[rlen] = '\0';

			if ((idr->flags[0] & 2) != 0 &&
			    ((rr_flags & RR_FLAG_CL) ||
			    (idr->name_len[0] != 1 ||
			    (idr->name[0] != 0 && idr->name[0] != 1)))) {
				/*
				 * This is a plain directory (neither "xxx/."
				 * nor "xxx/..").
				 * Add this directory to the todo list.
				 */
				int		dir_loop = 0;
				int		nextent;
				struct todo 	*tp = dp;

				stop_now = (*callback)(user_data, n, 1, 0, NULL);
				if (stop_now)
					break;

				nextent = isonum_733((unsigned char *)idr->extent);
				while (tp) {
					if (tp->extent == nextent) {
						dir_loop = 1;
						break;
					}
					tp = tp->prev;
				}
				if (dir_loop == 0) {
					td = (struct todo *) malloc(sizeof (*td));
					if (td == NULL)
						comerr(_("No memory.\n"));
					td->next = NULL;
					td->prev = dp;
					td->extent = nextent;
					td->length = isonum_733((unsigned char *)idr->size);
					td->name = (char *) malloc(strlen(rootname)
								+ strlen(name_buf) + 2);
					if (td->name == NULL)
						comerr(_("No memory.\n"));
					strcpy(td->name, rootname);
					strcat(td->name, name_buf);
					strcat(td->name, "/");

					*todo_pp = td;
					todo_pp = &td->next;
				}
			} else if ((idr->flags[0] & 2) &&
					(idr->name[0] != 0 || idr->name[0] != 1)) {
				/* this is . or .. */
			} else {
				extr_handle.dev = dev;
				extr_handle.idr = idr;
				stop_now = callback(user_data, n, 0,
						isonum_733((unsigned char *)idr->size), &extr_handle);
				if (stop_now)
					break;

				if (xtract && strcmp(xtract, n) == 0) {
					extract_file(dev, STDOUT_FILENO, idr, "stdout");
				}
			}
			if (do_f &&
			    (idr->name_len[0] != 1 ||
			    (idr->name[0] != 0 && idr->name[0] != 1))) {
				printf("%s\n", n);
			}
			if (Xtract) {
				/*
				 * In case if a multi-extent file, remember the
				 * start extent number.
				 */
				if ((idr->flags[0] & ISO_MULTIEXTENT) && size == 0)
					sextent = isonum_733((unsigned char *)idr->extent);

				if (debug ||
				    ((idr->flags[0] & ISO_MULTIEXTENT) == 0 && size == 0)) {
					if (dump_stat(rootname, idr, n,
							isonum_733((unsigned char *)idr->extent))) {
						if (Xtract) {
							if ((idr->flags[0] & 2) != 0 &&
							    idr->name_len[0] != 1 &&
							    idr->name[0] != 1) {
								char *p = n;
								if (*p == '/')
									p++;
								makedirs(p,
									S_IRUSR|S_IWUSR|S_IXUSR|
									S_IRGRP|S_IWGRP|S_IXGRP|
									S_IROTH|S_IWOTH|S_IXOTH,
									FALSE);
							} else {
								if (myuid != 0 &&
								    S_ISDIR(fstat_buf.st_mode)) {
									fstat_buf.st_mode |= S_IWUSR;
								}
								extract(dev, rootname, idr, n);
							}
						}
					}
				} else if (Xtract) {
					/*
					 * Extract all multi extent files here...
					 */
					extract(dev, rootname, idr, n);
				}
				size += fstat_buf.st_size;
				if ((flags & ISO_MULTIEXTENT) &&
				    (idr->flags[0] & ISO_MULTIEXTENT) == 0) {
					fstat_buf.st_size = size;
					if (!debug)
						idr->flags[0] |= ISO_MULTIEXTENT;
					dump_stat(rootname, idr, n, sextent);
					if (!debug)
						idr->flags[0] &= ~ISO_MULTIEXTENT;
				}
				flags = idr->flags[0];
				if ((idr->flags[0] & ISO_MULTIEXTENT) == 0)
					size = 0;
			}
		cont:
			i += buffer[i];
			if (i > 2048 - offsetof(struct iso_directory_record, name[0])) break;
		}

		if (stop_now)
			break;
	}
	if (Xtract) {
		char *nm = strrchr(rootname, '/');

		if (nm != rootname)
			nm++;
	}
	return stop_now;
}

EXPORT void extract_file_from_iso(context, out)
	struct iso_extract_handle *context;
	char *out;
{
	int		extent, len, tlen;
	int offs = 0;
	unsigned char	buff[20480];

	extent = isonum_733((unsigned char *)context->idr->extent);
	len = isonum_733((unsigned char *)context->idr->size);

	while (len > 0) {
		tlen = (len > sizeof (buff) ? sizeof (buff) : len);
#ifdef	USE_SCG
		readsecs(context->dev, extent - sector_offset, buff, ISO_BLOCKS(tlen));
#else
		lseek(fileno(infile), ((off_t)(extent - sector_offset)) << 11, SEEK_SET);
		read(fileno(infile), buff, tlen);
#endif
		memcpy(out + offs, buff, tlen);
		len -= tlen;
		offs += tlen;
		extent += ISO_BLOCKS(tlen);
	}
}

EXPORT int
scan_iso (filename, callback, user_data)
	const char* filename;
	iso_listing_callback callback;
	void* user_data;
{
	struct scsi_dev *dev;
	int retval;

	if (NULL == filename || NULL == callback)
		return ScanRetCode_BadInput;

	dev = scsidev_file_open(filename);
	retval = run_scanning(dev, callback, user_data);
	scsidev_close(dev);
	return retval;
}

EXPORT int
scan_device (sdevname, callback, user_data)
	const char* sdevname;
	iso_listing_callback callback;
	void* user_data;
{
#ifndef	USE_SCG
	return ScanRetCode_Disabled;
#else
	struct scsi_dev *dev;
	int retval;
	char* devname;

	if (NULL == callback)
		return ScanRetCode_BadInput;

	if (sdevname == NULL)
		cdr_defaults(&devname, NULL, NULL, NULL, NULL);
	if (devname == NULL)
		return ScanRetCode_BadInput;

	dev = scsidev_open(devname);
	if (scsidev_has_error(dev)) {
		retval = scsidev_error_code(dev);
		scsidev_close(dev);
		return retval;
	}

	retval = run_scanning(dev, callback, user_data);
	scsidev_close(dev);
	return retval;
#endif
}


LOCAL int
run_scanning(dev, callback, user_data)
	struct scsi_dev *dev;
	iso_listing_callback callback;
	void *user_data;
{
	int	c;
	int	ret = 0;
#if	defined(USE_NLS)
	char	*dir;
#endif
	/*
	 * Use toc_offset != 0 (-T #) if we have a complete multi-session
	 * disc that we want/need to play with.
	 * Here we specify the offset where we want to
	 * start searching for the TOC.
	 */
	int	toc_offset = 0;
	int	extent;
	struct todo * td;
	struct iso_primary_descriptor ipd;
	struct iso_primary_descriptor jpd;
	struct iso_directory_record * idr;
	char	*charset = NULL;
	int	voldesc_sum = 0;
	int stop_listing;
	char	*cp;

#if	defined(USE_NLS)
	setlocale(LC_ALL, "");
#if !defined(TEXT_DOMAIN)	/* Should be defined by cc -D */
#define	TEXT_DOMAIN "isoinfo"	/* Use this only if it weren't */
#endif
	dir = searchfileinpath("share/locale", F_OK,
					SIP_ANY_FILE|SIP_NO_PATH, NULL);
	if (dir)
		(void) bindtextdomain(TEXT_DOMAIN, dir);
	else
#if defined(PROTOTYPES) && defined(INS_BASE)
	(void) bindtextdomain(TEXT_DOMAIN, INS_BASE "/share/locale");
#else
	(void) bindtextdomain(TEXT_DOMAIN, "/usr/share/locale");
#endif
	(void) textdomain(TEXT_DOMAIN);
#endif

	myuid = getuid();

	charset = default_charset();

	if (charset == NULL) {
#if	(defined(__CYGWIN32__) || defined(__CYGWIN__) || defined(__DJGPP__) || defined(__MINGW32__)) && !defined(IS_CYGWIN_1)
		unls = sic_open("cp437");
#else
		unls = sic_open("default");
#endif
	} else {
		unls = sic_open(charset);
	}
	if (unls == NULL) {	/* Unknown charset specified */
		fprintf(stderr, _("Unknown charset: %s\nKnown charsets are:\n"),
							charset);
		sic_list(stdout); /* List all known charset names */

		list_locales();
		exit(EX_BAD);
		exit(1);
	}

	/*
	 * Absolute sector offset, so don't subtract sector_offset here.
	 */
#ifdef	USE_SCG
	readsecs(dev, 16 + toc_offset, &ipd, ISO_BLOCKS(sizeof (ipd)));
#else
	lseek(fileno(infile), ((off_t)(16 + toc_offset)) <<11, SEEK_SET);
	read(fileno(infile), &ipd, sizeof (ipd));
#endif
	idr = (struct iso_directory_record *)ipd.root_directory_record;
	for (c = 0, cp = (char *)&ipd; c < 2048; c++)
		voldesc_sum += *cp++ & 0xFF;
	if (GET_UBYTE(ipd.type) == ISO_VD_PRIMARY) {
		c = 17;
		do {
#ifdef	USE_SCG
			readsecs(dev, c + toc_offset, &jpd, ISO_BLOCKS(sizeof (jpd)));
#else
			lseek(fileno(infile), ((off_t)(c + toc_offset)) <<11, SEEK_SET);
			read(fileno(infile), &jpd, sizeof (jpd));
#endif
		} while (++c < 32 && GET_UBYTE(jpd.type) != ISO_VD_END);
		if (GET_UBYTE(jpd.type) == ISO_VD_END) do {
#ifdef	USE_SCG
			readsecs(dev, c + toc_offset, &jpd, ISO_BLOCKS(sizeof (jpd)));
#else
			lseek(fileno(infile), ((off_t)(c + toc_offset)) <<11, SEEK_SET);
			read(fileno(infile), &jpd, sizeof (jpd));
#endif
			cp = (char *)&jpd;
			if (strncmp(cp, "MKI ", 4) == 0) {
				int	sum;

				sum  = cp[2045] & 0xFF;
				sum *= 256;
				sum += cp[2046] & 0xFF;
				sum *= 256;
				sum += cp[2047] & 0xFF;
				if (sum == voldesc_sum)
					iso9660_inodes = TRUE;
				break;
			}
		} while (++c < 48);
	}

	/*
	 * Read '.' entry for the root directory.
	 */
	extent = isonum_733((unsigned char *)idr->extent);
#ifdef	USE_SCG
	readsecs(dev, extent - sector_offset, buffer, ISO_BLOCKS(sizeof (buffer)));
#else
	lseek(fileno(infile), ((off_t)(extent - sector_offset)) <<11, SEEK_SET);
	read(fileno(infile), buffer, sizeof (buffer));
#endif
	c = dump_rr(dev, (struct iso_directory_record *) buffer);
	if (c == 0 ||
	    (c & (RR_FLAG_SP | RR_FLAG_ER)) == 0 || su_version < 1 || rr_version < 1) {
		if (!debug)
			use_rock = FALSE;
	}
	/*
	 * ISO 9660:
	 *
	 *	DESC TYPE	== 1 (VD_PVD)	offset 0	len 1
	 *	STR ID		== "CD001"	offset 1	len 5
	 *	STD_VER		== 1		offset 6	len 1
	 */
	if ((ipd.type[0] != ISO_VD_PRIMARY) ||
	    (strncmp(ipd.id, ISO_STANDARD_ID, sizeof (ipd.id)) != 0) ||
	    (ipd.version[0] != 1)) {
		printf(_("CD-ROM is NOT in ISO 9660 format\n"));
		exit(1);
	}

	if (use_joliet) {
		int block = 16;
		movebytes(&ipd, &jpd, sizeof (ipd));
		while ((unsigned char) jpd.type[0] != ISO_VD_END) {
			if (debug && (unsigned char) jpd.type[0] == ISO_VD_SUPPLEMENTARY)
				error(_("Joliet escape sequence 0: '%c' 1: '%c' 2: '%c' 3: '%c'\n"),
					jpd.escape_sequences[0],
					jpd.escape_sequences[1],
					jpd.escape_sequences[2],
					jpd.escape_sequences[3]);
			/*
			 * Find the UCS escape sequence.
			 */
			if (jpd.escape_sequences[0] == '%' &&
			    jpd.escape_sequences[1] == '/' &&
			    (jpd.escape_sequences[3] == '\0' ||
			    jpd.escape_sequences[3] == ' ') &&
			    (jpd.escape_sequences[2] == '@' ||
			    jpd.escape_sequences[2] == 'C' ||
			    jpd.escape_sequences[2] == 'E')) {
				break;
			}

			block++;
#ifdef	USE_SCG
			readsecs(dev, block + toc_offset, &jpd, ISO_BLOCKS(sizeof (jpd)));
#else
			lseek(fileno(infile),
				((off_t)(block + toc_offset)) <<11, SEEK_SET);
			read(fileno(infile), &jpd, sizeof (jpd));
#endif
		}

		if (use_joliet && ((unsigned char) jpd.type[0] == ISO_VD_END)) {
			fprintf(stderr, _("Unable to find Joliet SVD\n"));
			exit(1);
		}

		switch (jpd.escape_sequences[2]) {
		case '@':
			ucs_level = 1;
			break;
		case 'C':
			ucs_level = 2;
			break;
		case 'E':
			ucs_level = 3;
			break;
		}

		if (ucs_level > 3) {
			fprintf(stderr,
				_("Don't know what ucs_level == %d means\n"),
				ucs_level);
			exit(1);
		}
		if (jpd.escape_sequences[3] == ' ')
			errmsgno(EX_BAD,
			_("Warning: Joliet escape sequence uses illegal space at offset 3\n"));
	}

	if (use_joliet)
		idr = (struct iso_directory_record *)jpd.root_directory_record;

	stop_listing = parse_dir(dev, todo_idr, "/",
				isonum_733((unsigned char *)idr->extent),
				isonum_733((unsigned char *)idr->size), callback, user_data);
	td = todo_idr;
	while (td) {
		if (!stop_listing)
			stop_listing = parse_dir(dev, td, td->name, td->extent, td->length, callback, user_data);
		free(td->name);
		td = td->next;
	}

	return (ret);
}

LOCAL void
list_locales()
{
	int	n;

	n = sic_list(stdout);
	if (n <= 0) {
		errmsgno(EX_BAD, "'%s/lib/siconv/' %s.\n",
			INS_BASE, n < 0 ? _("missing"):_("incomplete"));
		if (n == 0) {
			errmsgno(EX_BAD,
			_("Check '%s/lib/siconv/' for missing translation tables.\n"),
			INS_BASE);
		}
	}
#ifdef	USE_ICONV
	if (n > 0) {
		errmsgno(EX_BAD,
		_("'iconv -l' lists more available names.\n"));
	}
#endif
}

#include <schily/intcvt.h>

LOCAL int
time_cvt(dp, len)
	unsigned char	*dp;
	int		len;
{
	int	ret;

	for (ret = 0; --len >= 0; ) {
		ret *= 10;
		ret += *dp++ - '0';
	}
	return (ret);
}

LOCAL time_t
iso9660_time(date, hsecp, longfmt)
	unsigned char	*date;
	int		*hsecp;
	BOOL		longfmt;
{
	time_t	t;
	int	y;
	int	m;
	int	d;
	int	days;
	int	hour;
	int	min;
	int	sec;
	int	hsec;
	int	gmtoff;

	if (longfmt) {
		y = time_cvt(&date[0], 4);	/* Year 0..9999 */
		m = time_cvt(&date[4], 2);
		d = time_cvt(&date[6], 2);
		hour = time_cvt(&date[8], 2);
		min = time_cvt(&date[10], 2);
		sec = time_cvt(&date[12], 2);
		hsec = time_cvt(&date[14], 2);
		gmtoff = ((char *)date)[16];
	} else {
		y = date[0] + 1900;		/* Year 1900..2155 */
		m = date[1];
		d = date[2];
		hour = date[3];
		min = date[4];
		sec = date[5];
		hsec = 0;
		gmtoff = ((char *)date)[6];
	}
	if (hsecp)
		*hsecp = hsec;
	/*
	 * The original algorithm did win a Fortan contest in early times.
	 * It computes days relative to September 19th 1989.
	 * days = 367*(y-1980)-7*(y+(m+9)/12)/4-3*((y+(m-9)/7)/100+1)/4+275*m/9+d-100;
	 * The updated algorithm was modified to use Jan 1st 1970 as base
	 * and was taken from FreeBSD.
	 */
	days = 367*(y-1960)-7*(y+(m+9)/12)/4-3*((y+(m+9)/12-1)/100+1)/4+275*m/9+d-239;
	t = days;
	t = ((((t * 24) + hour) * 60 + min) * 60) + sec;
	if (-48 <= gmtoff && gmtoff <= 52)
		t -= gmtoff * 15 * 60;
	return (t);
}

LOCAL char*
default_charset(void)
{
	char *charset = NULL;

#if	defined(USE_NLS) && defined(HAVE_NL_LANGINFO) && defined(CODESET)
	/*
	 * If the locale has not been set up, nl_langinfo() returns the
	 * name of the default codeset. This should be either "646",
	 * "ISO-646", "ASCII", or something similar. Unfortunately, the
	 * POSIX standard does not include a list of valid locale names,
	 * so ne need to find all values in use.
	 *
	 * Observed:
	 * Solaris 	"646"
	 * Linux	"ANSI_X3.4-1968"	strange value from Linux...
	 */
	if (charset == NULL) {
		char	*codeset = nl_langinfo(CODESET);
		Uchar	*p;

		if (codeset != NULL)
			codeset = strdup(codeset);
		if (codeset == NULL)			/* Should not happen */
			goto setcharset;
		if (*codeset == '\0')			/* Invalid locale    */
			goto setcharset;

		for (p = (Uchar *)codeset; *p != '\0'; p++) {
			if (islower(*p))
				*p = toupper(*p);
		}
		p = (Uchar *)strstr(codeset, "ISO");
		if (p != NULL) {
			if (*p == '_' || *p == '-')
				p++;
			codeset = (char *)p;
		}
		if (strcmp("646", codeset) != 0 &&
		    strcmp("ASCII", codeset) != 0 &&
		    strcmp("US-ASCII", codeset) != 0 &&
		    strcmp("US_ASCII", codeset) != 0 &&
		    strcmp("USASCII", codeset) != 0 &&
		    strcmp("ANSI_X3.4-1968", codeset) != 0)
			charset = nl_langinfo(CODESET);

		if (codeset != NULL)
			free(codeset);

#define	verbose	0
		if (verbose > 0 && charset != NULL) {
			error(_("Setting input-charset to '%s' from locale.\n"),
				charset);
		}
	}
setcharset:
	/*
	 * Allow to switch off locale with -input-charset "".
	 */
	if (charset != NULL && *charset == '\0')
		charset = NULL;
#endif
	return charset;
}
